{----------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: OdbAdvan.Pas
|
| Description: Advantage database object database driver class.
|              For more information on the Advantage Database Server,
|              see http://www.advantagedatabase.com
|
| Notes: After running the automated tests with version 6 of the local
|        advantage engine:
|
|        Transactions are not supported, because Advantage does not support
|        them with binary memo fields
|
|
| History: Jun 13, 1998. Michel Brazeau, first version
|          Sep 15, 1998. Michel Brazeau. Tested with Delphi 4 and
|                        trial version 2.1 of ADS and there is a problem
|                        with autoincrement fields in InternalInsert.
|                        An exception with message "trying to modify a
|                        read only field" is raised, even when the
|                        autoinc field is not modified.
|          Sep 17, 1998. Michel Brazeau, received message from 
|                        Brett Donahue <brettd@extendsys.com> 
|                        indicating this problem will be fixed
|			 in the next release.
|          Oct 10, 2000. MB modifed for current ADS engine
|                        Add damSQL access method
|                        Only support ttAdsADT, to avoid dealing with
|                        XBase style indices and soft deleted records.
|          Mar 11, 2001. MB override introspection methods and testcase
|
|---------------------------------------------------------------------}
unit OdbAdvan;

{$I Odb.Inc}

interface

uses
    {$ifdef __Testing}
    DriverTest, { for TOdbDriverTest }
    {$endif}

    Classes,  { for TMemoryStream }
    SysUtils, { for Exception, StrPas }
    Db,       { for TBlobField, TDatabase }
    DbTables, { for TDatabase }
    AdsSet,   { for TAdsSettings }
    AdsTable, { for TAdsTable }
    AdsData,  { for TAdsTableTypes, ttAdsXXX }
    AdsCnnct, { for TAdsConnection }
    ObjList,  { for TUnOrderedList }
    Odb,      { for TObjectFile, TOdbDriver }
    OdbIndex, { for TOdbIndex }
    OdbField; { for TOdbField }

type

{--------------------------------------------------------------------}

TAdvantageOdbDriver = class(TOdbDriver)
protected

    FConnection         : TAdsConnection;

public

    constructor Create(const InDatabaseName : String); override;
    destructor Destroy; override;

    { MB Mar 11, 2001. From the ADE.Hlp file topic "TPS error conditions":

      Certain memo-related operations are invalid inside transactions such as
      updates of Blob fields or fields. For example: AdsFileToBinary and
      AdsSetBinary are illegal.

      This prevents using transactions with Spider ODB.

      Also as confirmed on Advantage.General newsgroup, transactions are not
      supported by the local engine. }

    procedure InternalOpenDatabase(
                            const ServerName : String;
                            const UserName   : String;
                            const Password   : String
                          ); override;

    procedure InternalCloseDatabase; override;

    function InternalOpenDataset( const ObjectDataset : TObjectDataset ) : TOdbHandle; override;

    function InternalOpenExclusive( const ObjectFile : TObjectFile ) : TOdbHandle; override;

    procedure InternalCloseDataset( const Handle : TOdbHandle); override;

    procedure InternalDeleteDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalCreateDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalInsert(
                      const Handle  : TOdbHandle;
                      const Fields  : TUnOrderedList;
                      const Stream  : TMemoryStream ); override;

    procedure InternalUpdate(
                      const Handle     : TOdbHandle;
                      const Fields     : TUnOrderedList;
                      const Stream     : TMemoryStream ); override;

    procedure InternalDelete( const Handle : TOdbHandle ); override;

    function InternalCreateObjectUpdate : TOdbHandle; override;
    procedure InternalDestroyObjectUpdate( const Handle : TOdbHandle ); override;

    procedure InternalInsertSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); override;

    procedure InternalUpdateSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); override;

    procedure InternalDeleteSQL( const Handle      : TOdbHandle;
                                 const ORMapping   : TORMapping;
                                 const DatasetName : String ); override;

    function InternalGetIndexName( const Handle : TOdbHandle ) : String; override;

    procedure InternalSetIndexName(
                            const Handle    : TOdbHandle;
                            const IndexName : String ); override;

    procedure InternalGetCurrent(
                          const Handle  : TOdbHandle;
                          const Stream  : TMemoryStream ); override;

    function InternalLockCurrent( const Handle : TOdbHandle ) : Boolean; override;

    procedure InternalUnLockCurrent( const Handle : TOdbHandle ); override;

    function InternalFindNext( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindPrior( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindFirst( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindLast( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindExact(
                        const Handle : TOdbHandle;
                        const Fields : TUnOrderedList;
                        const Index  : TOdbIndex ) : Boolean; override;

    procedure InternalFindNearest(
                          const Handle : TOdbHandle;
                          const Fields : TUnOrderedList;
                          const Index  : TOdbIndex ); override;

    function InternalGetBookmark( const Handle : TOdbHandle ) : Pointer; override;
    procedure InternalGotoBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;
    procedure InternalFreeBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;

    function InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt; override;

    function InternalIsEmpty( const Handle : TOdbHandle ) : Boolean; override;

    procedure InternalPack( const Handle : TOdbHandle ); override;

    function GetSupportedFieldTypes : TOdbFieldTypes; override;

    function GetAdsDataset( const Handle : TOdbHandle ) : TAdsDataset;

    function GetDataAccessMethods : TDataAccessMethodSet; override;

    function CheckDuplicateUpdate : Boolean; override;


    function AutoIncSupported              : Boolean; override;
    function IntegerIndexSupported         : Boolean; override;
    function FloatIndexSupported           : Boolean; override;
    function CaseInsensitiveIndexSupported : Boolean; override;
    function DescendingIndexSupported      : Boolean; override;
    function TransactionsSupported         : Boolean; override;
    function FileLockingSupported          : Boolean; override;
    function ObjectLockingSupported        : Boolean; override;
    function SingleIndexOnly               : Boolean; override;

protected
    FOwningComponent :  TComponent;

    function AdsOpenFile( const Name       : String;
                          const ObjectFile : TObjectFile;
                          const Exclusive  : Boolean ) : TOdbHandle;

    function AdsOpenQuery( const Name        : String;
                           const ObjectQuery : TObjectQuery ) : TOdbHandle;

    procedure OpenTableForObjectUpdate( const Handle      : TOdbHandle;
                                        const ORMapping   : TORMapping;
                                        const DatasetName : String );

end;

{$ifdef __Testing}

type

TAdvantageOdbDriverTest = class( TOdbDriverTest )
protected
     procedure GetParams( var DriverClass                   : TOdbDriverClass;
                          var DatabaseName                  : String;
                          var ServerName                    : String;
                          var UserName                      : String;
                          var Password                      : String
                        ); override;
end;

{$endif}

{--------------------------------------------------------------------}

implementation

uses
    FileCtrl, { for DirectoryExists }
    ObjPerst,
    ObjBuckt, { for TStringCombo }
    OdbMisc;  { for RaiseOdbException }

const

CObjectFieldName = 'Object';

var

{ MB Mar 11, 2001. Only one TAdsSettings per application }
AdsSettings     : TAdsSettings;

{--------------------------------------------------------------------}

constructor TAdvantageOdbDriver.Create(const InDatabaseName : String);
begin

    { Advantage does not allow creating a connection not owned by a
      component }
    FOwningComponent := TComponent.Create(nil);
    FConnection      := TAdsConnection.Create(FOwningComponent);

    if DirectoryExists(InDatabaseName) then
        FConnection.ConnectPath := InDatabaseName
    else
        FConnection.AliasName   := InDatabaseName;

    { the connection must be given a name, because Advantage uses the
      connection name to link a dataset with a connection }
    FConnection.Name := 'AdsConnection' + IntToStr(Random(99999));

    inherited Create(InDatabaseName);
end;

{--------------------------------------------------------------------}

destructor TAdvantageOdbDriver.Destroy;
begin
    { invoke inherited first, to close the database if it is opened before
      invoking FConnection.Free }
    inherited Destroy;

    FOwningComponent.Free;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalOpenDatabase(
                        const ServerName : String;
                        const UserName   : String;
                        const Password   : String
                      );
begin
    FConnection.IsConnected := True;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalCloseDatabase;
begin
    FConnection.IsConnected := False;
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.AdsOpenFile( const Name       : String;
                                          const ObjectFile : TObjectFile;
                                          const Exclusive  : Boolean ) : TOdbHandle;
var
    Table : TAdsTable;

begin
    Table := TAdsTable.Create(FOwningComponent);
    try
        Table.TableType    := ttAdsADT;
        Table.DatabaseName := FConnection.Name;
        Table.TableName    := Name;
        Table.Exclusive    := Exclusive;
        Table.Open;

        Result := NewHandle(Table);
    except
        Table.Free;
        raise;
    end;

end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.AdsOpenQuery(
                           const Name        : String;
                           const ObjectQuery : TObjectQuery ) : TOdbHandle;
var
    Query     : TAdsQuery;

begin
    Query     := nil;
    try
        Query     := TAdsQuery.Create(FOwningComponent);

        Query.DatabaseName := FConnection.Name;
        Query.SQL.Assign( ObjectQuery.Select );
        Query.Open;

        Result := NewHandle(Query);
    except
        Query.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalOpenDataset(
                       const ObjectDataset : TObjectDataset ) : TOdbHandle;
begin
    if ObjectDataset is TObjectFile then
        Result := AdsOpenFile( ObjectDataset.DatasetName,
                               TObjectFile(ObjectDataset), False)
    else if ObjectDataset is TObjectQuery then
    begin
        if TObjectQuery(ObjectDataset).Select.Count = 0 then
            RaiseOdbException( EOdbSelectStatementNotSpecified,
                               ObjectDataset.DatasetName );

        Result := AdsOpenQuery( ObjectDataset.DatasetName, TObjectQuery(ObjectDataset) );
    end
    else
    begin
        Result := 0; { prevent compiler warning }
        RaiseOdbException(EOdbError, 'Internal error ' + ObjectDataset.DatasetName);
    end;
end;
{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalOpenExclusive(
                                  const ObjectFile : TObjectFile ) : TOdbHandle;
begin
    Result := AdsOpenFile(ObjectFile.DatasetName, ObjectFile, True);
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalCloseDataset( const Handle : TOdbHandle);
var
    Dataset : TAdsDataset;
begin
    Dataset := GetAdsDataset(Handle);
    Dataset.Close;

    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalDeleteDataset( const ObjectDataset : TObjectDataset );
var
    Table : TAdsTable;
begin
    Table := TAdsTable.Create(FOwningComponent);
    try
        Table.TableType    := ttAdsADT;
        Table.DatabaseName := FConnection.Name;
        Table.TableName    := ObjectDataset.DatasetName;
        Table.Exclusive    := True;
        Table.DeleteTable;
    finally
        Table.Free;
    end;
end;

{--------------------------------------------------------------------}

{ add Fields to Table }
procedure AddFields( const Fields  : TUnOrderedList;
                     const Table   : TAdsTable );
var
    Field   : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftString  : Table.FieldDefs.Add(
                                 Field.Name,
                                 ftString, Field.Size-1, False );
            oftInteger : Table.FieldDefs.Add(
                                 Field.Name,
                                 ftInteger, 0, False );
            oftFloat   : Table.FieldDefs.Add(
                                 Field.Name,
                                 ftFloat, 0, False );
            {$ifndef Ver80}
            oftAutoInc : Table.FieldDefs.Add(
                                 Field.Name,
                                 ftAutoInc, 0, False );
            {$endif}

        end;
    until not Fields.GotoNext;

    Table.FieldDefs.Add(CObjectFieldName, ftBlob, 0, False);
end;

{--------------------------------------------------------------------}

{ convert Spider Odb index attributes to VCL index options }
function IndexAttributesToOptions(
            const Attributes : TOdbIndexAttributes ) : TIndexOptions;
begin
    Result := [];

    if iaUnique in Attributes then
        Result := Result + [ixUnique];

    if iaCaseInsensitive in Attributes then
        Result := Result + [ixCaseInsensitive];

    if iaDescending in Attributes then
        Result := Result + [ixDescending];

    if iaPrimary in Attributes then
        Result := Result + [ixPrimary];
end;

{--------------------------------------------------------------------}

procedure AddIndices( const Indices          : TUnOrderedList;
                      const FieldList        : TUnOrderedList;
                      const Table            : TAdsTable );
var
    Index        : TOdbIndex;
    IndexOptions : TIndexOptions;
    FieldNames   : String;
    IndexName    : String;

begin
    if Indices.GotoFirst then
    repeat
        Index        := Indices.CurrentObj as TOdbIndex;
        FieldNames   := Index.FieldNamesAsString(';');
        IndexOptions := IndexAttributesToOptions(Index.Attributes);

        if ixPrimary in IndexOptions then
            IndexOptions := IndexOptions - [ixPrimary] + [ixUnique];

        IndexName := Index.Name;

        Table.IndexDefs.Add( IndexName,
                             FieldNames,
                             IndexOptions );

    until not Indices.GotoNext;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalCreateDataset(
                          const ObjectDataset : TObjectDataset
                        );
var
    Table : TAdsTable;

begin
    Table := TAdsTable.Create(FOwningComponent);
    try
        Table.TableType    := ttAdsADT;
        Table.DatabaseName := FConnection.Name;
        Table.TableName    := ObjectDataset.DatasetName;
        Table.Exclusive    := True;

        Table.FieldDefs.Clear;
        Table.IndexDefs.Clear;

        AddFields(ObjectDataset.Fields, Table);
        AddIndices(ObjectDataset.Indices, ObjectDataset.Fields, Table);

        Table.CreateTable;
    finally
        Table.Free;
    end;
end;

{--------------------------------------------------------------------}

{ set the field values in Table given a list of TOdbField's }
procedure SetDatasetFieldValues( const Fields  : TUnOrderedList;
                                 const Dataset : TAdsDataset );
var
    Field : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftString  : Dataset.FieldByName(Field.Name).AsString  := Field.AsString;
            oftInteger : Dataset.FieldByName(Field.Name).AsInteger := Field.AsInteger;
            oftFloat   : Dataset.FieldByName(Field.Name).AsFloat   := Field.AsFloat;
        end;
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------}

{ set the field values in Table for the fields belonging to Index }
procedure SetTableFieldValuesOnIndex( const Fields  : TUnOrderedList;
                                      const Index   : TOdbIndex;
                                      const Table   : TAdsTable );
var
    Field       : TOdbField;
    StringCombo : TStringCombo;
begin
    if Index.FieldNames.GotoFirst then
    repeat
        StringCombo := Index.FieldNames.CurrentObj as TStringCombo;

        Field := GetFieldInList( StringCombo.Str, Fields );

        case Field.FieldType of
            oftString  : Table.FieldByName(Field.Name).AsString := Field.AsString;
            oftInteger,
            oftAutoInc : Table.FieldByName(Field.Name).AsInteger := Field.AsInteger;
            oftFloat   : Table.FieldByName(Field.Name).AsFloat := Field.AsFloat;
        end;
    until not Index.FieldNames.GotoNext;
end;

{--------------------------------------------------------------------}

{ Set the value on the autoincrement fields in Fields from the values
  contained in Table. }
procedure SetOdbAutoIncFieldValues( const Fields  : TUnOrderedList;
                                    const Table   : TAdsTable );
var
    Field : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftAutoInc : Field.AsInteger := Table.FieldByName(Field.Name).AsInteger;
        end;
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------}

procedure InsertInDataset( const Dataset : TAdsDataset;
                           const Fields  : TUnOrderedList;
                           const Stream  : TMemoryStream );
var
    Field   : TBlobField;
begin
    Dataset.Insert;
    try
        SetDatasetFieldValues(Fields, Dataset);

        Field := (Dataset.FieldByName(CObjectFieldName) as TBlobField);

        Field.LoadFromStream(Stream);

        Dataset.Post;

        { Update any autoincrement field in Fields. }
        if Dataset is TAdsTable then
             SetOdbAutoIncFieldValues(Fields, TAdsTable(Dataset));
    except
        Dataset.Cancel;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalInsert(
                  const Handle  : TOdbHandle;
                  const Fields  : TUnOrderedList;
                  const Stream  : TMemoryStream );
var
    Dataset : TAdsDataset;
begin
    Dataset := GetAdsDataset(Handle);

    InsertInDataset(Dataset, Fields, Stream);
end;

{--------------------------------------------------------------------}

procedure UpdateCurrentInDataset( const Dataset : TAdsDataset;
                                  const Fields  : TUnOrderedList;
                                  const Stream  : TMemoryStream );
begin
     try
        SetDatasetFieldValues(Fields, Dataset);
        (Dataset.FieldByName(CObjectFieldName) as TBlobField).LoadFromStream(Stream);
        Dataset.Post;
    except
        Dataset.Cancel;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalUpdate(
                  const Handle     : TOdbHandle;
                  const Fields     : TUnOrderedList;
                  const Stream     : TMemoryStream );
var
    Dataset       : TAdsDataset;
begin
    Dataset       := GetAdsDataset(Handle);

    try
        Dataset.Edit;
    except
        on Exception do
        begin
            if Dataset is TAdsTable then
            begin
                { attempt the unlock only when using TTable objects }
                InternalUnLockCurrent(Handle);
                Dataset.Edit;
            end
            else
                raise;
        end;
    end;

    UpdateCurrentInDataset(Dataset, Fields, Stream);
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalDelete( const Handle : TOdbHandle );
var
    Table : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;
    Table.Delete;

    { there is a glitch when using the ttAdsADT	table type when the
      last record is deleted. Positioning is not establish to the
      previous record as is with standard Delphi TTable behavior. }
    if Table.Bof and Table.Eof then
        Table.Last;
end;

{--------------------------------------------------------------------}

{$ifdef __Debug}
procedure SendDebugSQL( const SQL : TStrings );
var
    LineNo : Integer;
begin
    for LineNo := 0 to (SQL.Count-1) do
        SendDebug( SQL[LineNo] );
end;
{$endif}

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalCreateObjectUpdate : TOdbHandle;
var
    Dataset : TAdsTable;

begin
    Dataset := nil;
    try
        { MB Mar 23, 2001. Advantage does not support autoincrement fieldss
          with TQuery based updated, so a TTable is used }
        Dataset := TAdsTable.Create(nil);

        Dataset.DatabaseName := FConnection.Name;

        Result := NewHandle(Dataset);
    except
        Dataset.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalDestroyObjectUpdate( const Handle : TOdbHandle );
begin
    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.OpenTableForObjectUpdate(
                  const Handle      : TOdbHandle;
                  const ORMapping   : TORMapping;
                  const DatasetName : String );

var
    Table       : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;
    if not Table.Active then
    begin
        Table.TableName        := DatasetName;
        InternalSetIndexName(Handle, ORMapping.GetPrimaryIndex(DatasetName).Name);
        Table.Open;
    end;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalInsertSQL(
                  const Handle      : TOdbHandle;
                  const Stream      : TMemoryStream;
                  const ORMapping   : TORMapping;
                  const DatasetName : String );
begin
    OpenTableForObjectUpdate( Handle, ORMapping, DatasetName );

    InternalInsert(Handle, ORMapping.Fields, Stream );
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalUpdateSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String );
begin
    OpenTableForObjectUpdate( Handle, ORMapping, DatasetName );

    if not InternalFindExact( Handle, ORMapping.Fields, ORMapping.GetPrimaryIndex(DatasetName) ) then
        RaiseOdbException( EOdbObjectNotFound, 'TBdeOdbDriver.InternalUpdateSQL' );

    InternalUpdate(Handle, ORMapping.Fields, Stream );
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalDeleteSQL(
                             const Handle      : TOdbHandle;
                             const ORMapping   : TORMapping;
                             const DatasetName : String );
begin
    OpenTableForObjectUpdate( Handle, ORMapping, DatasetName );

    if not InternalFindExact( Handle, ORMapping.Fields, ORMapping.GetPrimaryIndex(DatasetName) ) then
        RaiseOdbException( EOdbObjectNotFound, 'TBdeOdbDriver.InternalDeleteSQL' );

    InternalDelete(Handle);
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalGetIndexName( const Handle : TOdbHandle ) : String;
var
    Table : TAdsTable;
begin
    Table  := GetAdsDataset(Handle) as TAdsTable;
    Result := Table.IndexName;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalSetIndexName(
                      const Handle    : TOdbHandle;
                      const IndexName : String );
var
    Table : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;

    Table.IndexName := IndexName;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalGetCurrent(
                      const Handle  : TOdbHandle;
                      const Stream  : TMemoryStream );
var
    Dataset : TAdsDataset;
    Field   : TBlobField;
begin
    Dataset := GetAdsDataset(Handle);

    Field := (Dataset.FieldByName(CObjectFieldName) as TBlobField);

    Field.SaveToStream(Stream);
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalLockCurrent( const Handle : TOdbHandle ) : Boolean;
var
    Table : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;

    Table.UpdateCursorPos;

    Result := Table.AdsLockRecord(0);
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalUnLockCurrent( const Handle : TOdbHandle );
var
    Table : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;

    Table.UpdateCursorPos;

    Table.AdsUnlockRecord(0);
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalFindNext( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TAdsDataset;
begin
    Dataset := GetAdsDataset(Handle);
    Dataset.Next;
    Result := not Dataset.EOF;
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalFindPrior( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TAdsDataset;
begin
    Dataset := GetAdsDataset(Handle);
    Dataset.Prior;
    Result := not Dataset.BOF;
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalFindFirst( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TAdsDataset;
begin
    Dataset := GetAdsDataset(Handle);
    Dataset.First;
    Result := not (Dataset.EOF and Dataset.BOF);
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalFindLast( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TAdsDataset;
begin
    Dataset := GetAdsDataset(Handle);
    Dataset.Last;
    Result := not (Dataset.EOF and Dataset.BOF);
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalFindExact(
                    const Handle : TOdbHandle;
                    const Fields : TUnOrderedList;
                    const Index  : TOdbIndex ) : Boolean;
var
    Table : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;
    Table.SetKey;
    SetTableFieldValuesOnIndex(Fields, Index, Table);
    Result := Table.GotoKey;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalFindNearest(
                      const Handle : TOdbHandle;
                      const Fields : TUnOrderedList;
                      const Index  : TOdbIndex );
var
    Table : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;
    Table.SetKey;
    SetTableFieldValuesOnIndex(Fields, Index, Table);
    Table.GotoNearest;
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalGetBookmark( const Handle : TOdbHandle ) : Pointer;
var
    Table : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;

    Result := Table.GetBookmark;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalGotoBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
var
    Table : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;

    Table.GotoBookmark(Bookmark);
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalFreeBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
var
    Table : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;

    Table.FreeBookmark(Bookmark);
end;


{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt;
begin
    Result := GetAdsDataset(Handle).RecordCount;
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.InternalIsEmpty( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TAdsDataset;
begin
    Dataset := GetAdsDataset(Handle);
    Result := Dataset.BOF and Dataset.EOF;
end;

{--------------------------------------------------------------------}

procedure TAdvantageOdbDriver.InternalPack( const Handle : TOdbHandle );
var
    Table : TAdsTable;
begin
    Table := GetAdsDataset(Handle) as TAdsTable;

    Table.AdsPackTable;
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.GetAdsDataset( const Handle : TOdbHandle ) : TAdsDataset;
begin
    Result := GetHandleDataObj(Handle) as TAdsDataset;
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.GetSupportedFieldTypes : TOdbFieldTypes;
begin
    Result := [oftString, oftInteger, oftFloat, oftAutoInc];
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.GetDataAccessMethods : TDataAccessMethodSet;
begin
    Result := [damISAM, damSQL];
end;

{--------------------------------------------------------------------}

function TAdvantageOdbDriver.CheckDuplicateUpdate : Boolean;
begin
    { MB Mar 11, 2000. Workaround a bug in Advantage engine. On a duplicate
      key update, an exception is raised, #7057 as expected, but the record is
      somehow updated. This was discovered with the automated tests,

      I have posted the following message on news 207.70.7.8/advantage.general:

       Subject:
            Problem with update on unique index
       Date:
            Sun, 11 Mar 2001 11:52:15 -0500
       From:
            Michel Brazeau <michel@intervalsoftware.com>
       Newsgroups:
            advantage.general

        Hi,

        I have just downloaded the local client version 6 engine. I am
        performing automated regression tests for the Spider Object Database.
        This is with the ttAdsADT table type.

        I have came across a problem when updating a record which violates a
        duplicate key. I have a simple unique string key, with values,

        'Object1', 'Object2', 'Object3', 'Object4', 'Object5', 'Object6',
        'Object7', 'Object8', 'Object9'

        I attempt to update 'Object1', with the value 'Object2', and the
        exception with code #7057 is properly raised, but then when I browse the
        data, my dataset is:

        'Object2', 'Object2', 'Object3', 'Object4', 'Object5', 'Object6',
        'Object7', 'Object8', 'Object9'

        I have two 'Object2', on a unique index.

        I am currently working around it by checking explicitley for duplicates
        on an update. Unfortunately my test program is too large to email this.
        I am quite sure of this behavior though. I run the same automated tests
        with DBISAM, BDE, and a Btrieve driver without problems.


        Another problem I came across is that transactions are not supported
        with memo binary fields. Is this a limitation of the local client engine
        only?
    }
    Result := True;
end;

{--------------------------------------------------------------------------}

function TAdvantageOdbDriver.AutoIncSupported              : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TAdvantageOdbDriver.IntegerIndexSupported         : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TAdvantageOdbDriver.FloatIndexSupported           : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TAdvantageOdbDriver.CaseInsensitiveIndexSupported : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TAdvantageOdbDriver.DescendingIndexSupported      : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TAdvantageOdbDriver.TransactionsSupported         : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TAdvantageOdbDriver.FileLockingSupported          : Boolean;
begin
    { see note in unit header }
    Result := False;
end;

{--------------------------------------------------------------------------}

function TAdvantageOdbDriver.ObjectLockingSupported        : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TAdvantageOdbDriver.SingleIndexOnly               : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------}

{$ifdef __Testing}

procedure TAdvantageOdbDriverTest.GetParams(
                    var DriverClass                   : TOdbDriverClass;
                    var DatabaseName                  : String;
                    var ServerName                    : String;
                    var UserName                      : String;
                    var Password                      : String
                  );
begin
    DriverClass  := TAdvantageOdbDriver;
    DatabaseName := Self.TestDbPath;
    ServerName   := '';
    UserName     := '';
    Password     := '';
end;

{$endif}

{--------------------------------------------------------------------}

initialization
    RegisterOdbDriver(TAdvantageOdbDriver, 'Advantage');

    AdsSettings  := TAdsSettings.Create(nil);

    { the default settings don't allow the local server }
    AdsSettings.AdsServerTypes := [stADS_REMOTE, stADS_LOCAL, stADS_AIS];

finalization

    AdsSettings.Free;

end.
