{---------------------------------------------------------------------------
|
|  Module: OdbIndex.Pas
|
|  Description: Database index abstraction classes.
|
|  History: Dec 31, 1997. Michel Brazeau, first version.
|           May 06, 1998. MB add IndexToXBaseIndexExp. Thanks to Yannick
|                         Skelling, skelling@integratik.com for the original
|                         code.
|           Feb   , 2001. MB Create unit tests for all routines
|
|---------------------------------------------------------------------------}
unit OdbIndex;

{$I Odb.Inc}

interface

uses
    {$ifdef __Testing}
    AutoTest, { for TSpiderOdbTestcase }
    {$endif}

    SysUtils, { for PString, NewStr, DisposeStr }
    ObjCmp,   { for TComparison, CompareStringNoCase }
    ObjList;  { for TUnOrderedList }

type

TOdbIndexAttribute    = ( iaPrimary,
                          iaUnique,
                          iaCaseInsensitive,
                          iaDescending
                        );

TOdbIndexAttributes = set of TOdbIndexAttribute;

type

TOdbIndex = class(TObject)
protected
    FFieldNames : TUnOrderedList;   { of TStringCombo }
    FAttributes : TOdbIndexAttributes;
    FName       : String;

public
    constructor Create( const IndexName    : String;
                        const InAttributes : TOdbIndexAttributes );
    destructor Destroy; override;

    property Attributes : TOdbIndexAttributes read FAttributes
                                           write FAttributes;

    { list of field names composing the index returned as a list of
      TStringCombo's. }
    property FieldNames : TUnOrderedList read FFieldNames;

    { returns the list of field names separated by Separator }
    function FieldNamesAsString( const Separator : Char ) : String;

    { returns True if the index has a field with a specific name }
    function HasField( const FieldName : String ) : Boolean;

    property Name : String read FName
                           write FName;
end;

function CompareOdbIndex(const Lhs, Rhs : TObject) : TComparison; far;

{ Generates an XBase index expression given an Index and a FieldList.
  IntegerSize is the Size of the integer field (Borland DBase BDE files
  use 11). FloatSize is the number of digits for floating point fields,
  and FloatDecimals is the number of decimals for floating point fields
  (Borland DBase BDE files use 20,4). }
function IndexToXBaseIndexExp( const Index         : TOdbIndex;
                               const FieldList     : TUnOrderedList;
                               const IntegerSize   : Integer;
                               const FloatSize     : Integer;
                               const FloatDecimals : Integer ) : String;

{$ifdef __Testing}

type

TOdbIndexUnitTest = class( TSpiderOdbTestCase )
published
     procedure TestOdbIndex;
     procedure TestCompareOdbIndex;
     procedure TestIndexToXBaseIndexExp;
end;

{$endif}

{--------------------------------------------------------------------}

implementation

uses
    ObjPerst, { for TObjectReader }
    ObjBuckt, { for TStringCombo }
    OdbField; { for TOdbField }


{--------------------------------------------------------------------------}

function CompareOdbIndex(const Lhs, Rhs : TObject) : TComparison;
var
    Left  : TOdbIndex absolute Lhs;
    Right : TOdbIndex absolute Rhs;
begin
    Result := CompareStringNoCase(Left.Name, Right.Name);
end;

{--------------------------------------------------------------------------}

constructor TOdbIndex.Create( const IndexName    : String;
                              const InAttributes : TOdbIndexAttributes );
begin
    inherited Create;

    FAttributes := InAttributes;
    FName       := IndexName;
    FFieldNames := TUnOrderedList.Create(TStringCombo, CompareStringComboNoCase);
end;

{---------------------------------------------------------------------------}

destructor TOdbIndex.Destroy;
begin
    FFieldNames.Free;

    inherited Destroy;
end;

{---------------------------------------------------------------------------}

function TOdbIndex.FieldNamesAsString( const Separator : Char ) : String;
var
    StringCombo : TStringCombo;
begin
    Result := '';

    if FieldNames.GotoFirst then
    repeat
        StringCombo := FieldNames.CurrentObj as TStringCombo;

        if Result = '' then
            Result := StringCombo.Str
        else
            Result := Result + Separator + StringCombo.Str;

    until not FieldNames.GotoNext;
end;

{---------------------------------------------------------------------------}

function TOdbIndex.HasField( const FieldName : String ) : Boolean;
var
    Key : TStringCombo;
begin
    Key := TStringCombo.Create( FieldName, nil );
    try
        Result := FieldNames.Search( Key, 1 );

    finally
        Key.Free;
    end;
end;

{---------------------------------------------------------------------------}

function IndexToXBaseIndexExp( const Index         : TOdbIndex;
                               const FieldList     : TUnOrderedList;
                               const IntegerSize   : Integer;
                               const FloatSize     : Integer;
                               const FloatDecimals : Integer ) : String;
var
    StringCombo : TStringCombo;
    Field       : TOdbField;
    Segment     : String;
begin
    Result := '';

    if Index.FieldNames.GotoFirst then
    repeat
        StringCombo := Index.FieldNames.CurrentObj as TStringCombo;
        Field       := GetFieldInList(StringCombo.Str, FieldList);

        case Field.FieldType of
            oftString  : Segment := StringCombo.Str;
            oftInteger : Segment := 'STR(' + StringCombo.Str + ',' +
                                    IntToStr(IntegerSize) + ',0)';
            oftFloat   : Segment := 'STR(' + StringCombo.Str + ',' +
                                    IntToStr(FloatSize) + ',' +
                                    IntToStr(FloatDecimals) + ')';
        end;

        if Result = '' then
            Result := Segment
        else
            Result := Result + '+' + Segment;

    until not Index.FieldNames.GotoNext;
end;

{---------------------------------------------------------------------------}

{$ifdef __Testing}

procedure TOdbIndexUnitTest.TestOdbIndex;
var
    Index     : TOdbIndex;
begin
    Index     := TOdbIndex.Create('Index', [iaPrimary, iaDescending] );
    try
        Index.FieldNames.Insert( TStringCombo.Create('IntFld1', nil ));
        Index.FieldNames.Insert( TStringCombo.Create('IntFld2', nil ));
        Index.FieldNames.Insert( TStringCombo.Create('FloatFld1', nil ));
        Index.FieldNames.Insert( TStringCombo.Create('StrFld1', nil ));

        Check( Index.Name = 'Index');
        Index.Name := 'Index1';
        Check( Index.Name = 'Index1');

        Check( Index.Attributes = [iaPrimary, iaDescending] );

        Index.Attributes := Index.Attributes + [iaCaseInsensitive];
        Check( Index.Attributes = [iaPrimary, iaDescending, iaCaseInsensitive] );

        Check( Index.FieldNamesAsString('+') = 'IntFld1+IntFld2+FloatFld1+StrFld1' );

        Check( Index.HasField( 'IntFld1' ) );
        Check( Index.HasField( 'FloatFld1' ) );
        Check( not Index.HasField( 'FloatFld2' ) );

    finally
        Index.Free;
    end;
end;

{---------------------------------------------------------------------------}

procedure TOdbIndexUnitTest.TestCompareOdbIndex;
var
    Index1  : TOdbIndex;
    Index2  : TOdbIndex;
begin
    Index1 := TOdbIndex.Create( 'Index1', [iaPrimary] );
    Index2 := TOdbIndex.Create( 'Index2', [iaPrimary] );
    try
        Check( CompareOdbIndex(Index1, Index2) = ObjCmp.cmpLess );
        Check( CompareOdbIndex(Index2, Index1) = ObjCmp.cmpGreater );

        Index2.Name := 'Index1';
        Check( CompareOdbIndex(Index2, Index1) = ObjCmp.cmpEqual );
    finally
        Index1.Free;
        Index2.Free;
    end;
end;

{---------------------------------------------------------------------------}

procedure TOdbIndexUnitTest.TestIndexToXBaseIndexExp;
var
    Index     : TOdbIndex;
    FieldList : TUnOrderedList;
begin
    Index     := TOdbIndex.Create('Index', [iaPrimary, iaDescending] );
    FieldList := TUnOrderedList.Create( TOdbField, CompareOdbField );
    try
        FieldList.Insert( TOdbField.Create('StrFld1', oftString, 21) );
        FieldList.Insert( TOdbField.Create('IntFld1', oftInteger, 99) );
        FieldList.Insert( TOdbField.Create('FloatFld1', oftFloat, 99) );
        FieldList.Insert( TOdbField.Create('StrFld2', oftString, 21) );
        FieldList.Insert( TOdbField.Create('IntFld2', oftInteger, 99) );
        FieldList.Insert( TOdbField.Create('FloatFld2', oftFloat, 99) );

        Index.FieldNames.Insert( TStringCombo.Create('IntFld2', nil ));
        Index.FieldNames.Insert( TStringCombo.Create('IntFld1', nil ));
        Index.FieldNames.Insert( TStringCombo.Create('FloatFld2', nil ));
        Index.FieldNames.Insert( TStringCombo.Create('StrFld1', nil ));

        Check( IndexToXBaseIndexExp( Index, FieldList, 10, 12, 6) =
                        'STR(IntFld2,10,0)+STR(IntFld1,10,0)+STR(FloatFld2,12,6)+StrFld1' );

    finally
        FieldList.Free;
        Index.Free;
    end;
end;

{$endif}

{---------------------------------------------------------------------------}

begin
    RegisterCompareFunction( 'CompareOdbIndex', CompareOdbIndex );
end.
